@@ -420,12 +420,21 @@ local sid_user = node.subtype 'user_defined'
 
 local prefix, inner_depth
 local utfchar = utf.char
+local function print_attr(p)
+--    local s= "      ==> "
+--    local n = p.attr.next
+--    while n do
+--        s = s .. ' [' .. tostring(n.number) .. ']=' .. tostring(n.value)
+--        n = node_next(n)
+--    end
+end
 local function debug_show_node_X(p,print_fn, limit, inner_depth)
    local k = prefix
    local s
-   local pt, pic = node_type(p.id), (get_attr(p, attr_icflag) or 0) % icflag_table.PROCESSED_BEGIN_FLAG
-   local base = prefix .. '[' .. string.format('%7d', node.direct.todirect(p)) .. '] ' ..
-     string.format('%X', pic) .. ' ' .. pt .. ' ' .. tostring(p.subtype) .. ' '
+   local pt, pic = node_type(p.id), (get_attr(p, attr_icflag) or 0) --% icflag_table.PROCESSED_BEGIN_FLAG
+   local base = prefix 
+--     .. string.format('[%9d] ', (node.direct.todirect(p)))
+     .. string.format('%4X', pic) .. ' ' .. pt .. ' ' .. tostring(p.subtype) .. ' '
    if pt == 'glyph' then
       s = base .. ' '
           .. (p.char<0xF0000 and utfchar(p.char) or '')
@@ -436,7 +445,7 @@ local function debug_show_node_X(p,print_fn, limit, inner_depth)
          s = s .. ' off: (' .. print_scaled(p.xoffset)
                .. ',' .. print_scaled(p.yoffset) .. ')'
       end
-      print_fn(s)
+      print_fn(s); print_attr(p)
    elseif pt=='hlist' or pt=='vlist' or pt=='unset'or pt=='ins' then
       if pt=='ins' then
          s = base .. '(' .. print_scaled(p.height) .. '+'
@@ -494,7 +503,7 @@ local function debug_show_node_X(p,print_fn, limit, inner_depth)
       end
       prefix=k
    elseif pt == 'glue' then
-      s = base .. ' ' ..  print_spec(p)
+      s = base .. print_spec(p)
       if pic>icflag_table.KINSOKU and pic<icflag_table.KANJI_SKIP then
          s = s .. ' (from JFM: priority ' .. pic-icflag_table.FROM_JFM .. ')'
       elseif pic==icflag_table.KANJI_SKIP then
@@ -508,10 +517,12 @@ local function debug_show_node_X(p,print_fn, limit, inner_depth)
       end
       print_fn(s)
    elseif pt == 'kern' then
-      s = base .. ' ' .. print_scaled(p.kern) .. 'pt'
-      if p.subtype==2 then
+      s = base .. print_scaled(p.kern) .. 'pt'
+      if p.subtype==0 then
+         s = s .. ' (font)'
+      elseif p.subtype==2 then
          s = s .. ' (for accent)'
-      elseif pic==icflag_table.IC_PROCESSED then
+      elseif p.subtype==3 or pic==icflag_table.IC_PROCESSED then
          s = s .. ' (italic correction)'
       elseif pic==icflag_table.LINEEND then
          s = s .. ' (end-of-line)'
@@ -519,7 +530,7 @@ local function debug_show_node_X(p,print_fn, limit, inner_depth)
          and pic<icflag_table.KANJI_SKIP then
          s = s .. ' (from JFM: priority ' .. pic-icflag_table.FROM_JFM .. ')'
       end
-      print_fn(s)
+      print_fn(s); print_attr(p)
    elseif pt == 'penalty' then
       s = base .. ' ' .. tostring(p.penalty)
       if pic==icflag_table.KINSOKU then s = s .. ' (for kinsoku)' end
@@ -532,10 +543,10 @@ local function debug_show_node_X(p,print_fn, limit, inner_depth)
          local t = tostring(p.user_id) .. ' (' ..
             luatexbase.get_user_whatsit_name(p.user_id) .. ') '
          if p.type ~= 110 then
-            s = s .. ' userid:' .. t .. tostring(p.value)
+            s = s .. 'userid:' .. t .. tostring(p.value)
             print_fn(s)
          else
-            s = s .. ' userid:' .. t .. '(node list)'
+            s = s .. 'userid:' .. t .. '(node list)'
             if p.user_id==uid_table.DIR then
                s = s .. ' dir: ' .. tostring(node.get_attribute(p, attr_dir))
             end
@@ -552,7 +563,7 @@ local function debug_show_node_X(p,print_fn, limit, inner_depth)
       else
          s = s .. (node.subtype(p.subtype) or '')
          if p.subtype==1 then
-            s = s .. ' stream=' .. p.stream
+            s = s .. 'stream=' .. p.stream
             print_fn(s)
             for i=1,#p.data do
                print_fn(s .. '  [' .. i .. '] = ' .. tostring(p.data[i] and p.date[i].csname))
