@@ -1,6 +1,6 @@
-# $Id: TLWinGoo.pm 63068 2022-04-18 05:58:07Z preining $
+# $Id: TLWinGoo.pm 72335 2024-09-21 09:54:18Z preining $
 # TeXLive::TLWinGoo.pm - Windows goop.
-# Copyright 2008-2021 Siep Kroonenberg, Norbert Preining
+# Copyright 2008-2024 Siep Kroonenberg, Norbert Preining
 # This file is licensed under the GNU General Public License version 2
 # or any later version.
 
@@ -13,7 +13,7 @@
 
 package TeXLive::TLWinGoo;
 
-my $svnrev = '$Revision: 63068 $';
+my $svnrev = '$Revision: 72335 $';
 my $_modulerevision;
 if ($svnrev =~ m/: ([0-9]+) /) {
   $_modulerevision = $1;
@@ -34,7 +34,6 @@ C<TeXLive::TLWinGoo> -- TeX Live Windows-specific support
 
 =head2 DIAGNOSTICS
 
-  TeXLive::TLWinGoo::is_ten;
   TeXLive::TLWinGoo::admin;
   TeXLive::TLWinGoo::non_admin;
   TeXLive::TLWinGoo::reg_country;
@@ -91,7 +90,6 @@ BEGIN {
   use vars qw( @ISA @EXPORT @EXPORT_OK $Registry);
   @ISA = qw( Exporter );
   @EXPORT = qw(
-    &is_ten
     &admin
     &non_admin
   );
@@ -179,36 +177,8 @@ if ($is_win) {
 
 =head2 DIAGNOSTICS
 
-=over 4
-
-=item C<win_version>
-
-C<win_version> returns the Windows version number as stored in the
-registry: 5.0 for Windows 2000, 5.1 for Windows XP and 6.0 for Vista.
-
 =cut
 
-my $windows_version = 0;
-my $windows_subversion = 0;
-
-if ($is_win) {
-  my $ver = `ver`;
-  chomp $ver;
-  $ver =~ s/^[^0-9]*//;
-  $ver =~ s/[^0-9.]*$//;
-  ($windows_version = $ver) =~ s/\..*$//;
-  ($windows_subversion = $ver) =~ s/^[^\.]*\.//;
-  $windows_subversion =~ s/\..*$//;
-}
-
-=item C<is_ten>
-
-C<is_ten> returns 1 if windows version is >= 10.0, otherwise 0.
-
-=cut
-
-sub is_ten { return $windows_version >= 10; }
-
 # permissions with which we try to access the system environment
 
 my $is_admin = 1;
@@ -493,6 +463,7 @@ sub adjust_reg_path_for_texlive {
   debug("Warning: [pdf]tex program not found in $tlbindir\n")
     if (!is_a_texdir($tlbindir));
   my $path = ($mode eq 'system') ? get_system_path() : get_user_path();
+  debug("TLWinGoo: adjust_reg_path_for_texlive: current path before adjusting = $path\n");
   $tlbindir =~ s!/!\\!g;
   my $tlbindir_short = uc(short_name($tlbindir));
   my ($d, $d_short, @newpath);
@@ -520,7 +491,7 @@ sub adjust_reg_path_for_texlive {
     }
   }
   if (@newpath) {
-    debug("TLWinGoo: adjust_reg_path_for_texlive: calling setenv_reg in $mode\n");
+    debug("TLWinGoo: adjust_reg_path_for_texlive: calling setenv_reg in $mode with new path = ", join(';', @newpath), "\n");
     setenv_reg("Path", join(';', @newpath), $mode);
   } else {
     debug("TLWinGoo: adjust_reg_path_for_texlive: calling unsetenv_reg in $mode\n");
@@ -1167,7 +1138,7 @@ sub create_uninstaller {
   my $td = $td_fw;
   $td =~ s!/!\\!g;
 
-  my $tdmain = `"$td\\bin\\win32\\kpsewhich" -var-value=TEXMFMAIN`;
+  my $tdmain = `"$td\\bin\\windows\\kpsewhich" -var-value=TEXMFMAIN`;
   $tdmain =~ s!/!\\!g;
   chomp $tdmain;
 
@@ -1182,7 +1153,7 @@ sub create_uninstaller {
   print UNINST <<UNEND;
 rem \@echo off
 setlocal
-path $td\\tlpkg\\tlperl\\bin;$td\\bin\\win32;%path%
+path $td\\tlpkg\\tlperl\\bin;$td\\bin\\windows;%path%
 set PERL5LIB=$td\\tlpkg\\tlperl\\lib
 rem Clean environment from other Perl variables
 set PERL5OPT=
@@ -1203,7 +1174,7 @@ set PERL_ROOT=
 set PERL_SIGNALS=
 set PERL_UNICODE=
 
-perl.exe \"$tdmain\\scripts\\texlive\\uninstall-win32.pl\" \%1
+perl.exe \"$tdmain\\scripts\\texlive\\uninstall-windows.pl\" \%1
 
 if errorlevel 1 goto :eof
 rem test for taskkill and try to stop exit tray menu
@@ -1242,7 +1213,7 @@ del /q \"$td\\release-texlive.txt\"
 UNEND2
 ;
   for my $d ('TEXMFSYSVAR', 'TEXMFSYSCONFIG') {
-    my $kd = `"$td\\bin\\win32\\kpsewhich" -var-value=$d`;
+    my $kd = `"$td\\bin\\windows\\kpsewhich" -var-value=$d`;
     chomp $kd;
     print UNINST2 "rmdir /s /q \"", $kd, "\"\r\n";
   }
@@ -1272,7 +1243,7 @@ UNEND3
   # but not for a user install under win10 because then
   # it shows up in Settings / Apps / Apps & features,
   # where it will trigger an inappropriate UAC prompt
-  if (admin() || !is_ten()) {
+  if (admin()) {
     &log("Registering uninstaller\n");
     my $k;
     my $uninst_key = $Registry -> Open((admin() ? "LMachine" : "CUser") .
